/*
 * This file is a part of the mg project.
 * Copyright (C) 1998 Martin Gall
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 */
/*
 *
 */

#include "config.h"

#ifdef HAVE_NIT
/* this source code is provided ``as is'' and has never been tested */
#include <sys/types.h>
#include <sys/socket.h>
#include <sys/ioctl.h>
#include <net/if.h>
#include <netinet/in.h>
#include <fcntl.h>
#include "pat_ether.h"
#include "pat_arp.h"
#include "pat_ip.h"
#include "pat_data.h"
#include "pat_ethaddr.h"
#include "pat_sz.h"
#include "xippkt.h"
#include "xipvar.h"

t_ether_addr		ethersun_ether_dst = {0x00,0x00,0xf8,0x00,0xb3,0x96};
char			ethersun_ifname[STR_BUFSIZ] = {'t','u','0',0};

int			ethersun_fd = -1;

/* opens named interface.
   Calls directly pfopen(2).
   Returns a file descriptor or a negative status meaning that something has 
   failed. */
int			ethersun_open(ifname)
char			*ifname;	/* Interface name */
{
  int			fd;
  struct ifreq		ifr;

  if (!ifname)
    {
      fprintf(stderr,"need interface name\n");
      exit(1);
    }
  if ((fd = open("/dev/nit",O_WRONLY)) < 0)
    {
      perror("open");
      exit(1);
    }
  strncpy(ifr.ifr_name,ifname,sizeof (ifr.ifr_name));
  if (ioctl(fd,NIOCBIND,&ifr) < 0)
    {
      perror("ioctl NIOCBIND");
      close(fd);
      exit(1);
    }
  return (fd);
}

/* close file descriptor */
VOID_FUNC		ethersun_close(fd)
int			fd;
{
  close(fd);
}

/* writes out an ethernet packet.
   Returns: See writev(2). 
   
   Warning: if dhost is incorrect (see etheraddr_from_str(3)) it
   might cause panic! */
int			ethersun_write(fd,dhost,type,databuf,datalen)
int			fd;		/* File descriptor */
t_ether_addr		*dhost;		/* Destination ethernet address */
int			type;		/* Host format of ethernet type */
char			*databuf;	/* Data buffer e.g starting
					   from ip header */
int			datalen;	/* Data length */
{
  struct sockaddr               sa;
  struct strbuf                 cbuf;
  struct strbuf                 dbuf;
  t_ether                       ether;

  sa.sa_family = AF_UNSPEC;
  FBCOPY(dhost,&(ether.dhost),sizeof (t_ether_addr));
  ether.type = htons(type);
  FBCOPY(&ether,sa.sa_data,sizeof (t_ether));
  cbuf.maxlen = cbuf.len = sizeof (sa);
  cbuf.buf = (char *)&sa;
  dbuf.len = pkt->len;
  dbuf.buf = pkt->buf;
  if (putmsg(fd,&cbuf,&dbuf,0) < 0)
    {
      perror("putmsg");
      exit(1);
    }
}

/* is a t_xip_method_proc.
   Sends out an ethernet packet.
   Is able to send "ether" pkts or either "ip" or "arp" in using
   global ethersun_ether_dst.
   Returns 0 if OK. -ERR_XIP_NO_METHOD if packet is not ether, ip or arp.
   Might return various errors */
t_status			ethersun_method_proc_send(xmd,args)
t_xip_method_data		*xmd;
t_vec				*args;
{
  t_ether_addr			dhost;
  int				type;
  t_pkt				sub_pkt;  
  t_pkt				*sent_pkt;
  t_status			status;
  t_in_addr			dst;
  char				buf[STR_BUFSIZ];

  if (xmd->xp->pkt->pat == &ether_pat)
    {
      buf[0] = 0;
      if ((status = pkt_get_field_to_str(xmd->xp->pkt,
					 "ether[0].dhost",
					 buf,
					 sizeof (buf))) < 0)
	return (status);
      if ((status = etheraddr_from_str(&dhost,
				       buf)) < 0)
	return (status);
      buf[0] = 0;
      if ((status = pkt_get_field_to_str(xmd->xp->pkt,
					 "ether[0].type",
					 buf,
					 sizeof (buf))) < 0)
	return (status);
      type = atoi(buf);
      if ((status = pkt_sub(xmd->xp->pkt,
			    &sub_pkt,
			    NULL)) < 0)
	return (status);
      sent_pkt = &sub_pkt;
    }
  else
    {
      if (xmd->xp->pkt->pat == &ip_pat)
	{
	  type = ETHERTYPE_IP;
	  buf[0] = 0;
	  if ((status = pkt_get_field_to_str(xmd->xp->pkt,
					     "ip[0].dst",
					     buf,
					     sizeof (buf))) < 0)
	    return (status);
	}
      else
	if (xmd->xp->pkt->pat == &arp_pat)
	  {
	    type = ETHERTYPE_ARP;
	    buf[0] = 0;
	    if ((status = pkt_get_field_to_str(xmd->xp->pkt,
					       "arpethip[0].tpa",
					       buf,
					       sizeof (buf))) < 0)
	      return (status);
	  }
	else
	  return (-ERR_XIP_NO_METHOD);
      sent_pkt = xmd->xp->pkt;
      bcopy((char *)(&ethersun_ether_dst),
	    (char *)(&dhost),
	    sizeof (t_ether_addr));
    }
  if (ethersun_write(ethersun_fd,
		       &dhost,
		       type,
		       sent_pkt->buf,
		       sent_pkt->len) < 0)
    return (-ERR_XIP_WRITE);
  return (0);
}

int	ethersunifname_chan;

PAT_NAME_GENERIC(ethersunifname_pat_name,
		 (VOID_PTR)&ethersunifname_chan,
		 "ethersunifname")

PAT_INSERT_DECL(ethersunifname_pat_insert)
{
  t_status	status;

  assert(sz_pat.insert_proc);
  if ((status = sz_pat.insert_proc(NULL,
				   buf,
				   len,
				   value)) < 0)
    return (status);
  if (ethersun_fd >= 0)
    ethersun_close(ethersun_fd);
  if ((ethersun_fd = ethersun_open(ethersun_ifname)) < 0)
    {
      err_print(ERR_XIP_OPEN,ethersun_ifname);
      return (0);
    }
#ifdef DEBUG
  if (XIP_VERB(VERB_XIP_MISC))
    fprintf(stderr,"ethersun_open %s, fd=%d\n",
	    ethersun_ifname,
	    ethersun_fd);
#endif
  return (0);
}

t_pat				ethersunifname_pat = 
{
  ethersunifname_pat_name,	/* t_pat_name_proc		*/
  NULL,				/* t_pat_off_proc		*/
  NULL,				/* t_pat_sub_proc		*/
  NULL,				/* t_pat_sum_proc		*/
  NULL,				/* t_pat_get_field_proc		*/
  NULL,				/* t_pat_set_field_proc		*/
  NULL,				/* t_pat_get_fields_proc	*/
  NULL,				/* t_pat_get_tmpl_proc		*/
  NULL,				/* t_pat_get_tmpl2_proc		*/
  NULL,				/* t_pat_has_opt_proc		*/
  NULL,				/* t_pat_adapt_len_proc		*/
  NULL,				/* t_pat_get_field_pat_proc	*/
  sz_pat_extract,		/* t_pat_extract_proc		*/
  ethersunifname_pat_insert,	/* t_pat_insert_proc		*/
  NULL,				/* t_pat_get_choices_proc	*/
};

t_xip_var		ethersun_vars[] = 
{
  {"ethersun_ether_dst",
   (t_off)(&ethersun_ether_dst),
   sizeof (ethersun_ether_dst),
   &etheraddr_pat,
   NULL},
  {"ethersun_ifname",
   (t_off)(&ethersun_ifname),
   sizeof (ethersun_ifname),
   &ethersunifname_pat,
   NULL},
  NULL_XIP_VAR
};

t_xip_cf_box_entry	ethersun_entries[] =
{
  {"ethersunEtherDst","ethersun_ether_dst"},
  {"ethersunIfName",	"ethersun_ifname"},
  NULL_XIP_CF_BOX_ENTRY
};

t_xip_cf_box_group	ethersun_group = 
{"ethersunGroup",	"ethersun",	ethersun_entries};

/* initializes sun stuff.
   Note: registers a new packet bar title, a new method and new variables.
   Returns 0. If it is not able to open ethersun_ifname (interface), it
   prints out an error message. */
t_status		xip_send_init(VOID_DECL)
{
  t_status		status;
  t_xip_method		xm;
  t_xip_pkt_bar		xpb;
  t_xip_var		*xv;
  
  xm.name = "send";
  xm.proc = ethersun_method_proc_send;
  if ((status = xip_method_add(&xm)) < 0)
    return (status);
  xpb.label = "Send";
  xpb.methodstr = "send()";
  if ((status = xip_pkt_bar_add(&xpb)) < 0)
    return (status);
  xv = ethersun_vars;
  while (xv->name)
    {
      if ((status = xip_var_add(xv)) < 0)
	return (status);
      xv++;
    }
  if ((status = xip_cf_box_group_add(&ethersun_group)) < 0)
    return (status);
  return (0);
}
#else
static int nit_kludge; /* avoid empty compilation */
#endif

