/*
 * This file is a part of the mg project.
 * Copyright (C) 1998 Martin Gall
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 */
/*
 *
 */

#include "config.h"

#if defined(HAVE_PF) || defined(HAVE_NIT)
int	sock_kludge; /* avoid empty compilation */
#else
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#ifdef HAVE_SOCK_PACKET
# include <linux/if_packet.h>
#endif
#include "pat.h"
#include "pat_rip.h"
#include "pat_ether.h"
#include "pat_ip.h"
#include "pat_arp.h"
#include "pat_ipproto.h"
#include "pat_boolean.h"
#include "xipvar.h"
#include "xippkt.h"

#define FAMILY_STR	"family="
#define PORT_STR	"port="
#define ADDR_STR	"addr="	
#define DEVICE_STR	"device="

/* converts a string representation to an opaque sockaddr.
   The correct definition string for a sockaddr_in is:
   family=inet;port=<port>;addr=<addr>[;].
   For a sockaddr_pkt (Linux) it will be:
   family=inet;device=<device>[;].
   Only this sockaddr types are implemented for the moment.
   Returns 0 if OK. Returns -ERR_XIP_BAD_SOCK_LEN if the value pointed by len
   is inferior to the desired sockaddr type. -ERR_BAD_SOCK_FAMILY is the
   family is not recognized. -ERR_NI if the family is recognized but not
   implemented. -ERR_XIP_BAD_SOCK_MEMBER if a member isn't recognized.
   Might return other various errors */
t_status		sa_from_str(sa,len,str,resolve)
struct sockaddr		*sa;	/* A opaque sockaddr structure */
int			*len;	/* Sockaddr len (filled) pointer */
char			*str;	/* Source string */
t_boolean		resolve;/* Try to resolve port and address */
{
  t_status		status;
  char			*str2;
  char			*limit;

  sa->sa_family = AF_UNSPEC;
  limit = str + strlen(str);
  str2 = str;
  while (str2 < limit &&
	 ((str = index(str2,';')) || 
	 ((str = index(str2,0)))))
    {
      t_assoc		*assoc;
      char		*value;
      char		*member;

      *str++ = 0;
      member = str2;
      str2 = str;
#ifdef NOTDEF
      printf("%s\n",member);
#endif
      switch (sa->sa_family)
	{
	case AF_UNSPEC:
	  if (*len < sizeof (struct sockaddr))
	    return (-ERR_XIP_BAD_SOCK_LEN);
	  *len = sizeof (struct sockaddr);
	  if (!strncmp(member,FAMILY_STR,strlen(FAMILY_STR)))
	    {
	      value = member + strlen(FAMILY_STR);
	      if (assoc = assoc_str_ptr_from_left(family_assocs,value))
		{
		  sa->sa_family = (int)(assoc->right);
		  continue ;
		}
	      else
		return (-ERR_XIP_BAD_SOCK_FAMILY);
	    }
	  break ;
	case AF_INET:
	  if (*len < sizeof (struct sockaddr_in))
	    return (-ERR_XIP_BAD_SOCK_LEN);
	  *len = sizeof (struct sockaddr_in);
	  if (!strncmp(member,PORT_STR,strlen(PORT_STR)))
	    {
	      value = member + strlen(PORT_STR);
	      if ((status = port_from_str(value,
					  "tcp",
					  resolve)) < 0)
		return (status);
	      ((struct sockaddr_in *)sa)->sin_port = htons(status);
	      continue ;
	    }
	  if (!strncmp(member,ADDR_STR,strlen(ADDR_STR)))
	    {
	      value = member + strlen(ADDR_STR);
	      if ((status = inaddr_from_str(value,
				&(((struct sockaddr_in *)sa)->sin_addr),
					    resolve)) < 0)
		return (status);
	      continue ;
	    }
#ifdef HAVE_SOCK_PACKET
	  if (!strncmp(member,DEVICE_STR,strlen(DEVICE_STR)))
	    {
	      value = member + strlen(DEVICE_STR);
	      ((struct sockaddr_pkt *)sa)->spkt_device[0] = 0;
	      if ((status = 
		   str_cat_str(((struct sockaddr_pkt *)sa)->spkt_device,
			   sizeof (((struct sockaddr_pkt *)sa)->spkt_device),
			       value)) < 0)
		return (status);
	      continue ;
	    }
#endif
	  break ;
	default:
	  return (-ERR_XIP_SOCK_NI);
	}
      return (-ERR_XIP_BAD_SOCK_MEMBER);
    }
  return (0);
}	

/* converts-and-catenates an opaque sockaddr to a string representation.
   It performs the contrary of sa_from_str(3).
   Returns 0 if OK. Returns -ERR_XIP_BAD_SOCK_LEN if len
   doesn't match the sockaddr family. Outputs a minimal representation
   for families recognized but not really implemented -ERR_NI otherwise */
t_status		sa_to_str(sa,len,str,max_len,resolve)
struct sockaddr		*sa;		/* An opaque sockaddr structure */
int			len;		/* Length of sockaddr */
char			*str;		/* A valid string */
int			max_len;	/* Maximum length */
t_boolean		resolve;	/* Try to resolve */
{
  t_assoc		*assoc;
  t_status		status;

  switch (sa->sa_family)
    {
    case AF_INET:
      if (len < sizeof (struct sockaddr_in))
	return (-ERR_XIP_BAD_SOCK_LEN);
      if (assoc = assoc_str_int_from_right(family_assocs,
					   sa->sa_family))
	{
	  if ((status = str_cat_fmt_va(str,
				       max_len,
				       "family=%s;",
				       assoc->left)) < 0)
	    return (status);
	}
      else
	{
	  if ((status = ulong_to_str((unsigned long)(sa->sa_family),
				     10,
				     str,
				     max_len)) < 0)
	    return (status);
	}
      if ((status = str_cat_str(str,
				max_len,
				"port=")) < 0)
	return (status);
      if ((status = port_to_str(ntohs(((struct sockaddr_in *)sa)->sin_port),
				"tcp",
				str,
				max_len,
				resolve)) < 0)
	return (status);
      if ((status = str_cat_str(str,
				max_len,
				";addr=")) < 0)
	return (status);
      return (inaddr_to_str(&(((struct sockaddr_in *)sa)->sin_addr),
			    str,
			    max_len,
			    resolve));
    default:
      if (len < sizeof (struct sockaddr))
	return (-ERR_XIP_BAD_SOCK_LEN);
      if (assoc = assoc_str_int_from_right(family_assocs,sa->sa_family))
	return (str_cat_fmt_va(str,
			       max_len,
			       "family=%s",
			       assoc->left));
      return (-ERR_XIP_SOCK_NI);
    }
  return (0);
}

t_assoc				socket_type_assocs[] = 
{
  {"stream",			(VOID_PTR)SOCK_STREAM},
  {"dgram",			(VOID_PTR)SOCK_DGRAM},
  {"raw",			(VOID_PTR)SOCK_RAW},
#ifdef HAVE_SOCK_PACKET
  {"packet",			(VOID_PTR)SOCK_PACKET},
#endif
  NULL_ASSOC
};

t_assoc				socket_flags_assocs[] = 
{
  {"oob",			(VOID_PTR)MSG_OOB},
  {"dontroute",			(VOID_PTR)MSG_DONTROUTE},
  {"none",			(VOID_PTR)0},
  {NULL,			NULL},
};

#ifdef HAVE_SOCK_PACKET
struct sockaddr_pkt		socket_addr = {
  AF_INET,
  {'e','t','h','0'}
};
t_u32				socket_type = SOCK_PACKET;
t_u32				socket_proto_ether_type = ETHERTYPE_IP;
#else
struct sockaddr			socket_addr = {AF_INET};
t_u32				socket_type = SOCK_RAW;
#endif

#ifdef HAVE_IP_HDRINCL
t_boolean			socket_ip_hdrincl = FALSE;
#endif

t_u32				socket_proto = IPPROTO_RAW;
t_u32				socket_flags = 0;
int				socket_addr_len = sizeof (socket_addr);

int				socket_fd = -1;

VOID_PTR			close_socket_if_open(VOID_DECL)
{
  if (socket_fd >= 0)
    {
      close(socket_fd);
      socket_fd = -1;
    }
}

VOID_PTR			reopen_socket(VOID_DECL)
{
  int				on;
  int				proto;
  int				family;

#ifdef HAVE_SOCK_PACKET
  if (socket_type == SOCK_PACKET)
    proto = socket_proto_ether_type;
  else
    proto = socket_proto;
#else
  proto = socket_proto;
#endif
#ifdef HAVE_SOCK_PACKET
  family = socket_addr.spkt_family;
#else
  family = socket_addr.sa_family;
#endif
  if ((socket_fd = socket(family,
			  socket_type,
			  proto)) < 0)
    {
      err_print(ERR_XIP_OPEN,"socket");
      return (0);
    }
#ifdef HAVE_IP_HDRINCL
  on = socket_ip_hdrincl;
  if (setsockopt(socket_fd,
		 IPPROTO_IP,
		 IP_HDRINCL,
		 (char *)&on,
		 sizeof (on)) < 0)
    {
      err_print(ERR_XIP_SYSCALL,"setsockopt IP_HDRINCL");
    }
#endif
#ifdef DEBUG
  if (XIP_VERB(VERB_XIP_MISC))
    fprintf(stderr,"socket fd=%d\n",socket_fd);
#endif
}

t_status			socket_method_proc_send(xmd,args)
t_xip_method_data		*xmd;
t_vec				*args;
{
  int				cc;

  if (xmd->xp->pkt->pat == &ip_pat && 
      socket_type == SOCK_RAW &&
      socket_proto == IPPROTO_RAW)
    {
      if ((cc = sendto(socket_fd,
		       xmd->xp->pkt->buf,
		       xmd->xp->pkt->len,
		       socket_flags,
		       (struct sockaddr *)&socket_addr,
		       socket_addr_len)) < 0)
	return (-ERR_XIP_WRITE);
      return (0);
    }
  else
    return (-ERR_XIP_NO_METHOD);
}

int	socketaddr_chan;

PAT_NAME_GENERIC(socketaddr_pat_name,
		 (VOID_PTR)&socketaddr_chan,
		 "socketaddr")

PAT_EXTRACT_DECL(socketaddr_pat_extract)
{
  return (sa_to_str(&socket_addr,
		    socket_addr_len,
		    str,
		    max_len));
}

PAT_INSERT_DECL(socketaddr_pat_insert)
{
  t_status		status;
  
  close_socket_if_open();
  if ((status = sa_from_str(&socket_addr,
			    &socket_addr_len,
			    value,
			    pat_resolve)) < 0)
    {
  
      err_print(-status,"sa_from_str");
      return (status);
    }
  reopen_socket();
  return (0);
}

t_pat				socketaddr_pat = 
{
  socketaddr_pat_name,		/* t_pat_name_proc		*/
  NULL,				/* t_pat_off_proc		*/
  NULL,				/* t_pat_sub_proc		*/
  NULL,				/* t_pat_sum_proc		*/
  NULL,				/* t_pat_get_field_proc		*/
  NULL,				/* t_pat_set_field_proc		*/
  NULL,				/* t_pat_get_fields_proc	*/
  NULL,				/* t_pat_get_tmpl_proc		*/
  NULL,				/* t_pat_get_tmpl2_proc		*/
  NULL,				/* t_pat_has_opt_proc		*/
  NULL,				/* t_pat_adapt_len_proc		*/
  NULL,				/* t_pat_get_field_pat_proc	*/
  socketaddr_pat_extract,	/* t_pat_extract_proc		*/
  socketaddr_pat_insert,	/* t_pat_insert_proc		*/
  NULL,				/* t_pat_get_choices_proc	*/
};

int	sockettype_chan;

PAT_NAME_GENERIC(sockettype_pat_name,
		 (VOID_PTR)&sockettype_chan,
		 "sockettype")

PAT_INSERT_DECL(sockettype_pat_insert)
{
  t_status		status;
  
  close_socket_if_open();
  if ((status = u32assoc_pat_insert(data,
				    buf,
				    len,
				    value)) < 0)
    {
      return (status);
    }
  reopen_socket();
  return (0);
}

t_pat				sockettype_pat = 
{
  sockettype_pat_name,		/* t_pat_name_proc		*/
  NULL,				/* t_pat_off_proc		*/
  NULL,				/* t_pat_sub_proc		*/
  NULL,				/* t_pat_sum_proc		*/
  NULL,				/* t_pat_get_field_proc		*/
  NULL,				/* t_pat_set_field_proc		*/
  NULL,				/* t_pat_get_fields_proc	*/
  NULL,				/* t_pat_get_tmpl_proc		*/
  NULL,				/* t_pat_get_tmpl2_proc		*/
  NULL,				/* t_pat_has_opt_proc		*/
  NULL,				/* t_pat_adapt_len_proc		*/
  NULL,				/* t_pat_get_field_pat_proc	*/
  u32assoc_pat_extract,		/* t_pat_extract_proc		*/
  sockettype_pat_insert,	/* t_pat_insert_proc		*/
  u32assoc_pat_get_choices,	/* t_pat_get_choices_proc	*/
};

int	socketproto_chan;

PAT_NAME_GENERIC(socketproto_pat_name,
		 (VOID_PTR)&socketproto_chan,
		 "socketproto")

PAT_INSERT_DECL(socketproto_pat_insert)
{
  t_status		status;
  
  close_socket_if_open();
  if ((status = ipproto_pat_insert(data,
				   buf,
				   len,
				   value)) < 0)
    {
      return (status);
    }
  reopen_socket();
  return (0);
}

t_pat				socketproto_pat = 
{
  socketproto_pat_name,		/* t_pat_name_proc		*/
  NULL,				/* t_pat_off_proc		*/
  NULL,				/* t_pat_sub_proc		*/
  NULL,				/* t_pat_sum_proc		*/
  NULL,				/* t_pat_get_field_proc		*/
  NULL,				/* t_pat_set_field_proc		*/
  NULL,				/* t_pat_get_fields_proc	*/
  NULL,				/* t_pat_get_tmpl_proc		*/
  NULL,				/* t_pat_get_tmpl2_proc		*/
  NULL,				/* t_pat_has_opt_proc		*/
  NULL,				/* t_pat_adapt_len_proc		*/
  NULL,				/* t_pat_get_field_pat_proc	*/
  ipproto_pat_extract,		/* t_pat_extract_proc		*/
  socketproto_pat_insert,	/* t_pat_insert_proc		*/
  ipproto_pat_get_choices,	/* t_pat_get_choices_proc	*/
};

int	socketprotoethertype_chan;

PAT_NAME_GENERIC(socketprotoethertype_pat_name,
		 (VOID_PTR)&socketprotoethertype_chan,
		 "socketprotoethertype")

PAT_INSERT_DECL(socketprotoethertype_pat_insert)
{
  t_status		status;
  
  close_socket_if_open();
  if ((status = u32assoc_pat_insert(data,
				    buf,
				    len,
				    value)) < 0)
    {
      return (status);
    }
  reopen_socket();
  return (0);
}

t_pat				socketprotoethertype_pat = 
{
  socketprotoethertype_pat_name,/* t_pat_name_proc		*/
  NULL,				/* t_pat_off_proc		*/
  NULL,				/* t_pat_sub_proc		*/
  NULL,				/* t_pat_sum_proc		*/
  NULL,				/* t_pat_get_field_proc		*/
  NULL,				/* t_pat_set_field_proc		*/
  NULL,				/* t_pat_get_fields_proc	*/
  NULL,				/* t_pat_get_tmpl_proc		*/
  NULL,				/* t_pat_get_tmpl2_proc		*/
  NULL,				/* t_pat_has_opt_proc		*/
  NULL,				/* t_pat_adapt_len_proc		*/
  NULL,				/* t_pat_get_field_pat_proc	*/
  nu32assoc_pat_extract,	/* t_pat_extract_proc		*/
  socketprotoethertype_pat_insert,/* t_pat_insert_proc		*/
  nu32assoc_pat_get_choices,	/* t_pat_get_choices_proc	*/
};

int	socketboolean_chan;

PAT_NAME_GENERIC(socketboolean_pat_name,
		 (VOID_PTR)&socketboolean_chan,
		 "socketboolean")

PAT_INSERT_DECL(socketboolean_pat_insert)
{
  t_status		status;
  
  close_socket_if_open();
  if ((status = boolean_pat_insert(data,
				   buf,
				   len,
				   value)) < 0)
    {
      return (status);
    }
  reopen_socket();
  return (0);
}

t_pat				socketboolean_pat = 
{
  socketboolean_pat_name,	/* t_pat_name_proc		*/
  NULL,				/* t_pat_off_proc		*/
  NULL,				/* t_pat_sub_proc		*/
  NULL,				/* t_pat_sum_proc		*/
  NULL,				/* t_pat_get_field_proc		*/
  NULL,				/* t_pat_set_field_proc		*/
  NULL,				/* t_pat_get_fields_proc	*/
  NULL,				/* t_pat_get_tmpl_proc		*/
  NULL,				/* t_pat_get_tmpl2_proc		*/
  NULL,				/* t_pat_has_opt_proc		*/
  NULL,				/* t_pat_adapt_len_proc		*/
  NULL,				/* t_pat_get_field_pat_proc	*/
  boolean_pat_extract,		/* t_pat_extract_proc		*/
  socketboolean_pat_insert,	/* t_pat_insert_proc		*/
  boolean_pat_get_choices,	/* t_pat_get_choices_proc	*/
};

t_xip_var		socket_vars[] = 
{
  {"socket_addr",
   (t_off)(&socket_addr),
   sizeof (socket_addr),
   &socketaddr_pat,
   NULL},
  {"socket_type",
   (t_off)(&socket_type),
   sizeof (socket_type),
   &sockettype_pat,
   socket_type_assocs},
  {"socket_proto",
   (t_off)(&socket_proto),
   sizeof (socket_proto),
   &socketproto_pat,
   NULL},
#ifdef HAVE_SOCK_PACKET
  {"socket_proto_ether_type",
   (t_off)(&socket_proto_ether_type),
   sizeof (socket_proto_ether_type),
   &socketprotoethertype_pat,
   ethertype_assocs},
#endif
  {"socket_flags",
   (t_off)(&socket_flags),
   sizeof (socket_flags),
   &u32assoc_pat,
   socket_flags_assocs},
#ifdef HAVE_IP_HDRINCL
  {"socket_ip_hdrincl",
   (t_off)(&socket_ip_hdrincl),
   sizeof (socket_ip_hdrincl),
   &socketboolean_pat,
   NULL},
#endif
  NULL_XIP_VAR
};

t_xip_cf_box_entry		socket_entries[] =
{
  {"socketAddr",		"socket_addr"},
  {"socketType",		"socket_type"},
  {"socketProto",		"socket_proto"},
#ifdef HAVE_SOCK_PACKET
  {"socketProtoEtherType",	"socket_proto_ether_type"},
#endif
  {"socketFlags",		"socket_flags"},
#ifdef HAVE_IP_HDRINCL
  {"socketIpHdrincl",		"socket_ip_hdrincl"},
#endif
  NULL_XIP_CF_BOX_ENTRY
};

t_xip_cf_box_group		socket_group = 
{"socketGroup",	"socket",	socket_entries};

/* initializes alpha stuff.
   Note: registers a new packet bar title, a new method and new variables.
   Returns 0. If it is not able to open socket_addr (interface), it
   prints out an error message. */
t_status		xip_send_init(VOID_DECL)
{
  t_status		status;
  t_xip_method		xm;
  t_xip_pkt_bar		xpb;
  t_xip_var		*xv;
  
  xm.name = "send";
  xm.proc = socket_method_proc_send;
  if ((status = xip_method_add(&xm)) < 0)
    return (status);
  xpb.label = "Send";
  xpb.methodstr = "send()";
  if ((status = xip_pkt_bar_add(&xpb)) < 0)
    return (status);
  xv = socket_vars;
  while (xv->name)
    {
      if ((status = xip_var_add(xv)) < 0)
	return (status);
      xv++;
    }
  if ((status = xip_cf_box_group_add(&socket_group)) < 0)
    return (status);
  return (0);
}
#endif
